{ NOTES:

  This program is written for the TURBO PASCAL compiler, version 4.0.
                     --------------------------
  The mathematics of this program is discussed elsewhere. From a programming
  point of the view the animation techniques are of interest.

  If "RECORD" is selected, then as each image is drawn, the screen
  area it occupies is copied to the "HEAP" The heap is an auxiliary
  storage area. Heap memory is allocated with the GETMEM command; it is
  NOT declared as a variable. Each call to GETMEM returns a POINTER to
  the allocated memory. We store these pointers in an array called "FRAME".

  On Playback, the images are transfered from the Heap to the screen area.
  (The procedure PLAY_BACK is very simple.)

  TURBO Pascal version 4 has useful routines for copying part of the screen
  to another area, usually the heap.  These routines include:
    IMAGESIZE
    GETIMAGE
    PUTIMAGE

  The amount of memory available on the heap is found through the function
    MEMAVAIL.
               ---------------------------------
 The program will work with most graphics adapters.
 If a VGA or EGA is installed, it will run in CGA 640x200 resolution.
 Higher resolution and extra colors are not needed here, and
 using EGA or VGA will slow the animation because of the need to
 transfer more memory.

 This is also the reason for keeping the size of the Hypercube small.

               ---------------------------------
 IF YOU ARE NEW TO PASCAL . . .

  This is a "Comment". Comments are for humans and are ignored by
  the Pascal Compiler. There are occaisonal comments throughout to
  explain some of the non-i=obvious operations.

  To "read" a PASCAL program, start from the end. In this case
  look for the line "MAIN PROGRAM STARTS HERE."

 }

 {U-}

  program HyperCube;
  Uses crt,Graph; {These are UNITS supplied with Turbo Pascal}
                  {The various graphics drivers, called BGI files
                   are included on this disk. They are distributed
                   with the permission of Borland International}


  const
   dimension = 4;
   number_of_verticies = 33;  {extra for return to first}
   max_frames = 300;          {for Animation, chosen arbitrarily,
                               number of frames actually limitted by
                               available memory on the HEAP}

type
    {real = extended;} {This type is faster in Turbo PASCAL 4.0
                       If you have an 80x87 chip, enable this
                       statement and recompile.}

    vertex = array[1..dimension] of real;
    rect = record {rectangle}
           left,top,right,bottom:integer;
          end; {record}
    string1 = string[1];
    string3 = string[3];
 var
    cube:array [1..number_of_verticies] of vertex;
    center_x,center_y:integer;
    X_scale,Y_scale:real;
    foreground_color,background_color:integer;
    key:char;
    rotation_keys:set of char;
    record_Images:boolean;
    frame:array[1..max_frames] of pointer;   {each will point to memory on HEAP}
    frame_number:integer;

     Display_Rect,
     frame_number_rect:rect; {locates display of frame number printout}
     Display_rect_size:longInt;
     Erase_needed:boolean;
     GraphDriver : integer;  { The Graphics Hardware device driver }

procedure beep;
 begin
  sound(440);
  delay(300);
  nosound;
 end;


procedure SetRect(var Rectangle:rect; l,t,r,b:integer);
 begin
   with rectangle do
    begin
     left := l;
     right := r;
     top := t;
     bottom := b;
    end;
 end;


procedure RectBorder(R:Rect);
 {sames as the RECTANGLE routine, but different input type}
 begin
   setcolor(1);
   with R do
     rectangle(left,top,right,bottom);
 end;

procedure EraseRect(Var R:rect);
 begin
 setcolor(0);
  SetFillStyle(solidfill,0); {0 is background color}
  with r do
   bar3d(r.left,r.top,r.right,r.bottom,0,false);
 end;

procedure CenterText(Y:integer; S:string);
 {Centers text within Display_rect, at given y-position}
 var x:integer;
  begin
   with Display_rect do
     begin
       x := (left+right-textWidth(s)) div 2;
       OutTextXY(x,y,s);
     end
  end;

procedure Draw_key(left,top:integer; c:string1; axes:string3);
 const edge = 20;
 begin
   rectangle(left,top,left+2*edge,top+edge);
   outTextXY(left+edge div 2+5, top + 2,c);
   outTextXY(left+edge div 2, top + 11,axes);
 end;

procedure Draw_Command_key(left,top:integer; c,s:string);
 const
  edge = 20;
 var
  R:rect;
 begin
   setRect(R,left,top,left+6 * edge,top+edge);
   eraseRect(R);
   rectborder(r);
   outTextXY(left+3*edge-textwidth(c) Div 2 , top + 2,c);
   outTextXY(left+edge div 2, top + 11,s);
 end;



procedure Hide_playback_key;
 begin
   with Display_rect do
    draw_Command_key(left + 2 * 130 ,bottom+20, 'C', '');
 end;

procedure Show_playback_key;
 begin
 with Display_rect do
   draw_Command_key(left + 2 * 130 ,bottom+20, 'C', '  Play Back  ');
 end;

procedure Hide_Record_key;
 begin
   with Display_rect do
    draw_Command_key(left + 1 * 130 ,bottom + 20, 'X', '');
 end;


procedure Show_Record_key;
 begin
   with Display_rect do
    draw_Command_key(left + 1 * 130 ,bottom + 20, 'X', 'Start Record');
 end;



 procedure Draw_keys;
   var
    i:integer;
    s1:string1;
    s3:string3;
   BEGIN
    for i := 1 to 6 do
     begin
       s1:=copy('123456',i,1);
       s3 := copy('x-yx-zy-zx-ty-tz-t',1+3*(i-1),3);
       draw_key(display_rect.right+ 20+ i*45, 50, s1, s3);

       s1:=copy('QWERTY',i,1);
       s3 := copy('y-xz-xz-yt-xt-yt-z',1+3*(i-1),3);
       draw_key(display_rect.right + 30+i*45, 75, s1, s3);
     end;
     outtextXY(display_rect.right + 80,100,'Rotation keys');
    with Display_rect do
      begin
       draw_Command_key(left + 0 * 130 ,bottom + 20, 'Z', '  Reset Cube');
       draw_Command_key(left + 0 * 130 ,5, 'Esc', 'End Program!');
       show_record_key;
       show_playback_key;
       draw_Command_key(left + 3 * 130 ,bottom + 20, 'V', '   Lascaux');
      end;
   END; {draw keys}




 procedure Init_graphics;    {See note on graphics drivers above}
  var
    
    GraphMode   : integer;  { The Graphics mode value }
    errorcode  : integer;
   begin
       DetectGraph(GraphDriver, GraphMode);
       case GraphDriver of
          {        vga: graphmode := vgahi;} {VGA owners, who would like to
                                              take full advantage of the
                                              adapter, can enable this branch
                                              of the case statement. You will
                                              also have to fiddle with the scales
                                              and positions elsewhere in the
                                              program. We recommend you do
                                              this only if you have a FAST
                                              machine with lots of memory.}
          CGA,EGA,VGA:begin
                        GRAPHDRIVER := CGA;
                        Graphmode := CGAHi;
                      end;
                  HercMono:begin
                            GRAPHDRIVER := HercMono;
                            Graphmode := HercMonoHi;
                           end;
                  else begin
                         writeln('Don''t recognize graphics hardware.');
                         readln;
                      end;
       end; {case}



     INITGRAPH(GraphDriver,GRAPHMODE,'C:\TP');  {!!!!!!}
     ErrorCode := graphresult;
     if ErrorCode <> grOk then
       begin
            Writeln('Graphics error: ', GraphErrorMsg(ErrorCode));
            Halt(1);
        end;
   end;



procedure initialize;
 const
   edge = 50;
 var
   I:integer;
  begin
   foreground_color := white;
   background_color := black;
   {
     With CGA's we assume High Resolution, so correct for skewed
     proportions between Horizontal and veritcal,

     GetMaxX, GetMaxY are Turbo 4 functions that return the
     screen size.
   }
 CASE GRAPHDRIVER OF
  VGA,EGA,CGA: SetRect(display_rect, 50,getMaxY div 2 - edge,
                 50 + 4*edge ,getMaxY div 2 + edge);
     HercMono: SetRect(display_rect, 50,getMaxY div 2 - 2*edge,
                 50 + 5*edge ,getMaxY div 2 + 2*edge);
 END;

    with display_rect do
     begin
       Display_Rect_Size := ImageSize(left,top,right,bottom);
       {Turbo Pascal function returns number of bytes in screen image.
	used for the animation.}

        {put frame number bellow display rect}
       setRect(frame_number_rect,left,top -20 ,left+4*edge,top -1);
       center_x := (left+right) div 2; {of display_rect}
       center_y := (top+bottom) div 2;
     end;

 CASE GRAPHDRIVER OF
  VGA,EGA,CGA: begin
                 Y_scale := center_y * 0.2;
                 X_scale := 2*y_scale;
               end;
     HercMono: begin
                 Y_scale := center_y * 0.3;
                 X_scale := 1.2*y_scale;
               end;
 END;
     record_images := false;
     Erase_needed := false;
     frame_number := 1;

     for i := 1 to max_frames do
      frame[i] := nil; {no memory allocated get}
  end; {Initialize}

procedure initialize_cube;
  begin
    cube[1,1] :=  1; cube[1,2] :=  1; cube[1,3] :=  1;      cube[1,4] := 1;
    cube[2,1] :=  1; cube[2,2] :=  1; cube[2,3] :=  1;      cube[2,4] := -1;
    cube[3,1] :=  1; cube[3,2] :=  1; cube[3,3] := -1;      cube[3,4] :=-1;
    cube[4,1] :=  1; cube[4,2] := -1; cube[4,3] := -1;      cube[4,4] :=-1;
    cube[5,1] := -1; cube[5,2] := -1; cube[5,3] := -1;      cube[5,4] :=-1;
    cube[6,1] := -1; cube[6,2] := -1; cube[6,3] := -1;      cube[6,4] := 1;
    cube[7,1] := -1; cube[7,2] := -1; cube[7,3] :=  1;      cube[7,4] := 1;
    cube[8,1] := -1; cube[8,2] :=  1; cube[8,3] :=  1;      cube[8,4] := 1;
    cube[9,1] :=  1; cube[9,2] :=  1; cube[9,3] :=  1;      cube[9,4] := 1;
    cube[10,1] :=  1; cube[10,2] :=  1; cube[10,3] := -1;   cube[10,4] := 1;
    cube[11,1] :=  1; cube[11,2] := -1; cube[11,3] := -1;   cube[11,4] := 1;
    cube[12,1] :=  1; cube[12,2] := -1; cube[12,3] :=  1;   cube[12,4] := 1;
    cube[13,1] :=  1; cube[13,2] := -1; cube[13,3] :=  1;   cube[13,4] :=-1;
    cube[14,1] := -1; cube[14,2] := -1; cube[14,3] :=  1;   cube[14,4] :=-1; 
    cube[15,1] := -1; cube[15,2] :=  1; cube[15,3] :=  1;   cube[15,4] :=-1; 
    cube[16,1] := -1; cube[16,2] :=  1; cube[16,3] := -1;   cube[16,4] :=-1;
    cube[17,1] := -1; cube[17,2] :=  1; cube[17,3] := -1;   cube[17,4] := 1;
 
    cube[18,1] := -1; cube[18,2] :=  1; cube[18,3] :=  1;   cube[18,4] := 1;
    cube[19,1] := -1; cube[19,2] :=  1; cube[19,3] :=  1;   cube[19,4] :=-1;
    cube[20,1] :=  1; cube[20,2] :=  1; cube[20,3] :=  1;   cube[20,4] :=-1; 
    cube[21,1] :=  1; cube[21,2] := -1; cube[21,3] :=  1;   cube[21,4] :=-1; 
    cube[22,1] :=  1; cube[22,2] := -1; cube[22,3] := -1;   cube[22,4] :=-1; 
    cube[23,1] :=  1; cube[23,2] := -1; cube[23,3] := -1;   cube[23,4] := 1;
    cube[24,1] := -1; cube[24,2] := -1; cube[24,3] := -1;   cube[24,4] := 1;
    cube[25,1] := -1; cube[25,2] :=  1; cube[25,3] := -1;   cube[25,4] := 1; 

    cube[26,1] :=  1; cube[26,2] :=  1; cube[26,3] := -1;   cube[26,4] := 1;
    cube[27,1] :=  1; cube[27,2] :=  1; cube[27,3] := -1;   cube[27,4] :=-1;
    cube[28,1] := -1; cube[28,2] :=  1; cube[28,3] := -1;   cube[28,4] :=-1;
    cube[29,1] := -1; cube[29,2] := -1; cube[29,3] := -1;   cube[29,4] :=-1; 
    cube[30,1] := -1; cube[30,2] := -1; cube[30,3] :=  1;   cube[30,4] :=-1;
    cube[31,1] := -1; cube[31,2] := -1; cube[31,3] :=  1;   cube[31,4] := 1;
    cube[32,1] :=  1; cube[32,2] := -1; cube[32,3] :=  1;   cube[32,4] := 1;
    cube[33,1] :=  1; cube[33,2] :=  1; cube[33,3] :=  1;   cube[33,4] := 1;

  end;


procedure  multiply_all_verticies(axis_1,axis_2,sign:integer);
  {This is essentially a Matrix Multiplying each vertex}
      var
       angle:real;
       s,c:real;
       which_vertex,j:integer;
       temp:vertex;
      begin
        angle := pi/24;        {step size is arbitrary      }
        s := sin(sign*angle);  {Can save a few microseconds }
        c := cos(sign*angle);  {by making these global and  }
                               {computing them only once.   }

        for which_vertex:= 1 to number_of_verticies do
            begin
              temp := cube[which_vertex];
              temp[axis_1] := c* cube[which_vertex,axis_1] - s * cube[which_vertex,axis_2];
              temp[axis_2] := s* cube[which_vertex,axis_1] + c * cube[which_vertex,axis_2];
              cube[which_vertex] :=temp;
            end;
      end;

procedure Display_frame_number;
  var
   fns:string;
 begin
  eraseRect(Frame_number_rect);
  moveto(frame_number_rect.left+10,frame_number_rect.top+2);
  str(frame_number,FNS);
  setcolor(1);
  OutText('Frame # '); OutText(fns);
  str(MemAvail div display_rect_size,FNS);{number of remaining frames}
  OutText('  ('+Fns+' left)');
end;


procedure Turn_on_record;
begin
  record_images := true;
  Display_frame_number;
  draw_Command_key(display_rect.left + 1 * 130 ,display_rect.bottom + 20, 'X', ' Stop Record');
  {put blanks in label for PLAYBACK key}
  Hide_playback_key;
end;

  procedure Turn_off_Record;
   begin
    frame_number:= frame_number-1;
    record_images := false;
    eraseRect(frame_number_rect);
    show_record_key;
    show_playback_key;
    end;

procedure Reset_frames;
 var i:integer;
  begin
   for i := 1 to max_frames do     {release memory if was allocated before}
     if frame[i] <> nil
       then FreeMem(frame[i],display_rect_size);
   frame_number := 1;
  end;

procedure draw_cube(color:integer);
    var
       i:integer;
       x,y:integer;
       Grab_Ok:boolean;
    begin
     If Erase_Needed {only after a PLAY_BACK has occured}
       then begin
              EraseRect(display_rect);
              RectBorder(display_Rect);
            end;
     Erase_needed := false;
     SetColor(color);
     if (Record_images) and (color = 0)
         then begin
                display_frame_number;
                setcolor(color);
                   {check if there is enough Heap storage space.}
                Grab_Ok :=  (maxavail > Display_rect_size +100); {100 is arbitrary safety margin}
                if Grab_ok
                  then begin
                        GetMem(frame[frame_number],display_rect_size); {allocate on HEAP}
                        with Display_rect do
                        begin
                         GetImage(left,top,right,bottom, {transfer image to HEAP}
                                  frame[frame_number]^ );
                         frame_number := frame_number+1;
                         end
                       end
                   else begin
                          beep;
                          turn_off_Record;
                          eraseRect(display_rect);
                          rectborder(display_rect);
                         end;
               end;

             {project onto x-y plane, (center_X, center_Y, scales are set elsewhere)}

        x := trunc(center_X + X_scale * cube[1,1]); {can pre-multiply to save a few microSecs}
        y := trunc(center_Y - Y_scale * cube[1,2]);
        MoveTo (x,y);
        for i := 2 to number_of_verticies do
           begin
              x := trunc(center_X + X_scale * cube[i,1]);
              y := trunc(center_Y - Y_scale * cube[i,2]);
              LineTo (x,y);
           end;

    end; {draw cube}

procedure LascauX;
 var
  hold_image :pointer;
  Enough_Mem:boolean;
  K:char;
 begin

    with Display_rect do
     begin
      Enough_mem :=  (maxavail > Display_rect_size +100); {100 is arbitrary safety margin}
      if enough_mem
         then begin
               GetMem(hold_image,display_rect_size);
               GetImage(left,top,right,bottom,hold_image^ );
             end;
      eraseRect(display_rect);
      rectborder(display_rect);
      CenterText(top+10,'Lascaux Graphics');
      CenterText(top+30,'Software for Mathematics');
      CenterText(top+50,'3220 Steuben Ave,');
      CenterText(top+70,'Bronx, NY  10467');
      CenterText(top+90,'(212) 654-7429');
      {outTextXY(left+10,top+10,'   Lascaux Graphics');
      outTextXY(left+5,top+30,'Software for Mathematics');
      outTextXY(left+5,top+50,'    3220 Steuben Ave,');
      outTextXY(left+5,top+70,'    Bronx, NY  10467');
      outTextXY(left+10,top+90,'    (212) 654-7429');
      }
      repeat until Keypressed; {just wait}
       k:= readkey; {throw away result of keyboard read}
      If enough_mem
         then begin
               putImage(left,top,hold_image^,normalPut);
               freeMem(hold_image,display_rect_size);
             end
     end;
  end;


  procedure Play_back;
   var i:integer;
    begin
      if frame_number>1
      then
       for i := 1 to frame_number do
        begin
          with display_rect do
           putImage(left,top,frame[i]^,normalPut);
       end;
       Erase_needed:= true;  {Before next draw. Necessary because
                             we only play back the images, not the
                             actual coordinates. Regular erase may
                             not erase an image drawn by PLAY_BACK}
    end;


   procedure Main_Loop;
     var
      itemHit:integer;
      k:char;
      axis_1,axis_2,sign:integer;
      border_rect:rect;

     begin
         {Disable "play back" button, for now}
         draw_Cube(foreground_color);
         rectborder(display_rect);
         rotation_keys := ['1','q','Q',
                           '2','w','W',
                           '3','e','E',
                           '4','r','R',
                           '5','t','T',
                           '6','y','Y'];
              Repeat
                gotoxy(1,1);
                k:=readkey;
                if k in Rotation_keys
                   then draw_Cube(background_color);{erase previous}
                case k of
                    '1' : begin
                                     axis_1 := 1;
                                     axis_2 := 2;
                                     sign := 1;
                                   end;
                    'q','Q': begin
                        axis_1 := 1;
                        axis_2 := 2;
                       sign := -1;
                      end;

                    '2': begin
                        axis_1 := 1;
                        axis_2 := 3;
                       sign := 1;
                      end;
                    'w','W': begin
                        axis_1 := 1;
                        axis_2 := 3;
                       sign := -1;
                      end;
                    '3': begin
                        axis_1 := 2;
                        axis_2 := 3;
                       sign := 1;
                      end;
                    'e','E': begin
                        axis_1 := 2;
                        axis_2 := 3;
                       sign := -1;
                      end;
                    '4': begin
                        axis_1 := 1;
                        axis_2 := 4;
                       sign := 1;
                      end;
                    'r','R': begin
                        axis_1 := 1;
                        axis_2 := 4;
                       sign := -1;
                      end;
                    '5': begin
                        axis_1 := 2;
                        axis_2 := 4;
                       sign := 1;
                      end;
                    't','T': begin
                        axis_1 := 2;
                        axis_2 := 4;
                       sign := -1;
                      end;
                    '6': begin
                        axis_1 := 3;
                        axis_2 := 4;
                       sign := 1;
                      end;
                    'y','Y': begin
                        axis_1 := 3;
                        axis_2 := 4;
                       sign := -1;
                      end;

          'x','X': if not record_images  {interpert button as record on}
                             then begin
                                   reset_Frames;
                                   turn_on_record;
                                  end
                             else begin
                                    Turn_off_record;
                                  end;
                   'c','C': if not record_images
                                then begin
                                      Hide_record_key;
                                      PLAY_BACK;
                                      SHow_record_key;
                                    end;
                        'v','V': Lascaux; {our message}
                        'z','Z': Begin {reset cube to original position}
                                  draw_Cube(background_color);{erase previous}
                                  Initialize_cube;
                                  draw_Cube(foreground_color);
                                 end;
                  end; {case}
                  if k in rotation_keys
                    then begin
                          multiply_all_verticies(axis_1,axis_2,sign);
                          draw_cube(foreground_color);
                          if record_images and (frame_number = max_frames)
                              then begin  {force STOP RECORD if limit exceeded}
                                    beep;
                                    turn_off_record;
                                   end;
                         end;
                 until K = #27; {Esc = QUIT}
     end; {CUBE_DIALOG}



begin  {Main program starts here.}
  Init_graphics;
  initialize;
  initialize_cube;         {1's and -1's in verticies }
  Draw_keys;
  Main_Loop;
  restoreCrtMode;
 END.

